@@ -38,17 +38,16 @@ struct BezierCurve
   void render(const triple *p, GLuint I0, GLuint I1);
   
   void append() {
-    material1Data.append1(data);
-    
-    if(material1Data.vertices1.size() >= gl::maxvertices) {
-      drawBuffer(material1Data,noNormalShader);
-      material1Data.clear();
-      gl::forceRemesh=true;
-    }
+    material1Data.append(data);
   }
   
+  void notRendered() {
+    material1Data.rendered=false;
+  }
+
   void queue(const triple *g, bool straight, double ratio) {
     data.clear();
+    notRendered();
     Onscreen=true;
     init(pixel*ratio);
     render(g,straight);
@@ -62,12 +61,6 @@ struct Pixel
   
   void append() {
     material0Data.append0(data);
-    
-    if(material0Data.vertices0.size() >= gl::maxvertices) {
-      drawBuffer(material0Data,pixelShader);
-      material0Data.clear();
-      gl::forceRemesh=true;
-    }
   }
   
   void queue(const triple& p, double width) {
