@@ -18,24 +18,55 @@ protected:
   const path3 g;
   triple center;
   bool straight;
-  prc::RGBAColour color;
+  prc::RGBAColour diffuse;
+  prc::RGBAColour emissive;
+  prc::RGBAColour specular;
+  double opacity;
+  double shininess;
+  double metallic;
+  double fresnel0;
   bool invisible;
   Interaction interaction;
   triple Min,Max;
+  bool billboard;
+  size_t centerIndex;  
 public:
 #ifdef HAVE_GL
-  static BezierCurve R;
+  BezierCurve R;
 #endif  
-  drawPath3(path3 g, triple center, const pen& p, Interaction interaction,
-            const string& key="") :
-    drawElement(key), g(g), center(center), straight(g.piecewisestraight()),
-    color(rgba(p)), invisible(p.invisible()), interaction(interaction),
-    Min(g.min()), Max(g.max()) {}
+  void init() {
+#ifdef HAVE_LIBOSMESA
+    billboard=false;
+#else
+    billboard=interaction == BILLBOARD;
+#endif
+    centerIndex=0;
+  }
+  
+  drawPath3(path3 g, triple center, const vm::array& p, double opacity,
+            double shininess, double metallic, double fresnel0,
+            Interaction interaction, const string& key="") :
+    drawElement(key), g(g), center(center),
+    straight(g.piecewisestraight()), opacity(opacity),
+    shininess(shininess), metallic(metallic), fresnel0(fresnel0),
+    interaction(interaction), Min(g.min()), Max(g.max()) {
+    init();
+
+    pen Pen=vm::read<camp::pen>(p,0);
+    invisible=Pen.invisible();
+    diffuse=rgba(Pen);
+    emissive=rgba(vm::read<camp::pen>(p,1));
+    specular=rgba(vm::read<camp::pen>(p,2));
+  }
     
   drawPath3(const double* t, const drawPath3 *s) :
     drawElement(s->KEY), g(camp::transformed(t,s->g)), straight(s->straight),
-    color(s->color), invisible(s->invisible), interaction(s->interaction),
+    diffuse(s->diffuse), emissive(s->emissive), specular(s->specular),
+    opacity(s->opacity), shininess(s->shininess),
+    metallic(s->metallic), fresnel0(s->fresnel0),
+    invisible(s->invisible), interaction(s->interaction),
     Min(g.min()), Max(g.max()) {
+    init();
     center=t*s->center;
   }
   
@@ -68,10 +99,16 @@ public:
     } else b=pair(m(b.getx(),z.getx()),m(b.gety(),z.gety()));
   }
   
+  void meshinit() {
+    if(billboard)
+      centerIndex=centerindex(center);
+  }
+  
   bool write(prcfile *out, unsigned int *, double, groupsmap&);
+  bool write(jsfile *out);
   
   void render(double, const triple&, const triple&, double,
-              bool transparent);
+              bool remesh);
 
   drawElement *transformed(const double* t);
 };
@@ -87,7 +124,7 @@ protected:
   bool invisible;
   triple Min,Max;
   
-#ifdef HAVE_GL
+#ifdef HAVE_LIBGLM
   GLfloat *Controls;
   GLfloat *Knots;
 #endif  
@@ -126,7 +163,7 @@ public:
     
     run::copyArrayC(knots,knot,0,NoGC);
     
-#ifdef HAVE_GL
+#ifdef HAVE_LIBGLM
     Controls=NULL;
 #endif  
   }
@@ -138,7 +175,7 @@ public:
     for(unsigned int i=0; i < n; ++i)
       controls[i]=t*s->controls[i];
     
-#ifdef HAVE_GL
+#ifdef HAVE_LIBGLM
     Controls=NULL;
 #endif    
   }
@@ -156,7 +193,7 @@ public:
              bool &first);
     
   void render(double size2, const triple& Min, const triple& Max,
-              double perspective, bool transparent);
+              double perspective, bool remesh);
     
   drawElement *transformed(const double* t);
 };
@@ -171,7 +208,7 @@ class drawPixel : public drawElement {
   triple Min,Max;
 public:
 #ifdef HAVE_GL
-  static Pixel R;
+  Pixel R;
 #endif  
   drawPixel(const triple& v, const pen& p, double width, const string& key="")
     : drawElement(key), v(v), p(p), color(rgba(p)), width(width),
@@ -194,9 +231,10 @@ public:
   }
   
   void render(double size2, const triple& b, const triple& B,
-              double perspective, bool transparent);
+              double perspective, bool remesh);
   
   bool write(prcfile *out, unsigned int *, double, groupsmap&);
+  bool write(jsfile *out);
   
   drawElement *transformed(const double* t);
 };
