@@ -31,6 +31,7 @@ size_t drawElement::lastcenterIndex=0;
 const triple drawElement::zero;
 
 using vm::array;
+using settings::getSetting;
 
 #ifdef HAVE_LIBGLM
 
@@ -58,18 +59,11 @@ void setcolors(bool colors,
                const RGBAColour& specular, double shininess,
                double metallic, double fresnel0, jsfile *out)
 {
-  Material m;
-  if(colors) {
-    static glm::vec4 Black(0.0,0.0,0.0,diffuse.A);
-    m=Material(Black,Black,
-               glm::vec4(specular.R,specular.G,specular.B,specular.A),
-               shininess,metallic,fresnel0);
-  } else
-    m=Material(glm::vec4(diffuse.R,diffuse.G,diffuse.B,diffuse.A),
-               glm::vec4(emissive.R,emissive.G,emissive.B,emissive.A),
-               glm::vec4(specular.R,specular.G,specular.B,specular.A),
-               shininess,metallic,fresnel0);
-          
+  Material m=Material(glm::vec4(diffuse.R,diffuse.G,diffuse.B,diffuse.A),
+                      glm::vec4(emissive.R,emissive.G,emissive.B,emissive.A),
+                      glm::vec4(specular.R,specular.G,specular.B,specular.A),
+                      shininess,metallic,fresnel0);
+  
   MaterialMap::iterator p=materialMap.find(m);
   if(p != materialMap.end()) materialIndex=p->second;
   else {
@@ -89,7 +83,7 @@ void drawBezierPatch::bounds(const double* t, bbox3& b)
 {
   double x,y,z;
   double X,Y,Z;
-  
+
   if(straight) {
     triple Vertices[4];
     if(t == NULL) {
@@ -103,48 +97,48 @@ void drawBezierPatch::bounds(const double* t, bbox3& b)
       Vertices[2]=t*controls[12];
       Vertices[3]=t*controls[15];
     }
-    
+
     boundstriples(x,y,z,X,Y,Z,4,Vertices);
   } else {
     double cx[16];
     double cy[16];
     double cz[16];
-  
+
     if(t == NULL) {
-      for(int i=0; i < 16; ++i) {
+      for(unsigned int i=0; i < 16; ++i) {
         triple v=controls[i];
         cx[i]=v.getx();
         cy[i]=v.gety();
         cz[i]=v.getz();
       }
     } else {
-      for(int i=0; i < 16; ++i) {
+      for(unsigned int i=0; i < 16; ++i) {
         triple v=t*controls[i];
         cx[i]=v.getx();
         cy[i]=v.gety();
         cz[i]=v.getz();
       }
     }
-    
+
     double c0=cx[0];
     double fuzz=Fuzz*run::norm(cx,16);
     x=bound(cx,min,b.empty ? c0 : min(c0,b.left),fuzz,maxdepth);
     X=bound(cx,max,b.empty ? c0 : max(c0,b.right),fuzz,maxdepth);
-    
+
     c0=cy[0];
     fuzz=Fuzz*run::norm(cy,16);
     y=bound(cy,min,b.empty ? c0 : min(c0,b.bottom),fuzz,maxdepth);
-    Y=boundtri(cy,max,b.empty ? c0 : max(c0,b.top),fuzz,maxdepth);
-    
+    Y=bound(cy,max,b.empty ? c0 : max(c0,b.top),fuzz,maxdepth);
+
     c0=cz[0];
     fuzz=Fuzz*run::norm(cz,16);
     z=bound(cz,min,b.empty ? c0 : min(c0,b.near),fuzz,maxdepth);
     Z=bound(cz,max,b.empty ? c0 : max(c0,b.far),fuzz,maxdepth);
-  }  
-  
+  }
+
   b.add(x,y,z);
   b.add(X,Y,Z);
-  
+
   if(t == NULL) {
     Min=triple(x,y,z);
     Max=triple(X,Y,Z);
@@ -207,11 +201,11 @@ void drawBezierPatch::ratio(const double* t, pair &b, double (*m)(double, double
 
 bool drawBezierPatch::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
-  if(invisible || !prc)
+  if(invisible || primitive)
     return true;
 
   RGBAColour Black(0.0,0.0,0.0,diffuse.A);
-  PRCmaterial m(Black,diffuse,emissive,specular,opacity,PRCshininess);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
 
   if(straight) {
     triple vertices[]={controls[0],controls[12],controls[3],controls[15]};
@@ -229,7 +223,7 @@ bool drawBezierPatch::write(prcfile *out, unsigned int *, double, groupsmap&)
 bool drawBezierPatch::write(jsfile *out)
 {
 #ifdef HAVE_LIBGLM
-  if(invisible)
+  if(invisible || primitive)
     return true;
 
   if(billboard) {
@@ -239,11 +233,13 @@ bool drawBezierPatch::write(jsfile *out)
   
   setcolors(colors,diffuse,emissive,specular,shininess,metallic,fresnel0,out);
   
+  out->precision(digits);
   if(straight) {
     triple Controls[]={controls[0],controls[12],controls[15],controls[3]};
     out->addPatch(Controls,4,Min,Max,colors,4);
   } else
     out->addPatch(controls,16,Min,Max,colors,4);
+  out->precision(getSetting<Int>("digits"));
                     
 #endif  
   return true;
@@ -257,20 +253,6 @@ void drawBezierPatch::render(double size2, const triple& b, const triple& B,
   transparent=colors ? colors[0].A+colors[1].A+colors[2].A+colors[3].A < 4.0 :
     diffuse.A < 1.0;
   
-  bool offscreen;
-  if(billboard) {
-    drawElement::centerIndex=centerIndex;
-    BB.init(center);
-    offscreen=bbox2(Min,Max,BB).offscreen();
-  } else
-    offscreen=bbox2(Min,Max).offscreen();
-  
-  if(offscreen) { // Fully offscreen
-    S.Onscreen=false;
-    S.data.clear();
-    return;
-  }
-
   setcolors(colors,diffuse,emissive,specular,shininess,metallic,fresnel0);
   
   if(transparent)
@@ -282,6 +264,21 @@ void drawBezierPatch::render(double size2, const triple& b, const triple& B,
       setMaterial(materialData,drawMaterial);
   }
   
+  bool offscreen;
+  if(billboard) {
+    drawElement::centerIndex=centerIndex;
+    BB.init(center);
+    offscreen=bbox2(Min,Max,BB).offscreen();
+  } else
+    offscreen=bbox2(Min,Max).offscreen();
+  
+  if(offscreen) { // Fully offscreen
+    S.Onscreen=false;
+    S.data.clear();
+    S.notRendered();
+    return;
+  }
+
   triple *Controls;
   triple Controls0[16];
   if(billboard) {
@@ -366,7 +363,7 @@ void drawBezierTriangle::bounds(const double* t, bbox3& b)
         cz[i]=v.getz();
       }
     }
-    
+
     double c0=cx[0];
     double fuzz=Fuzz*run::norm(cx,10);
     x=boundtri(cx,min,b.empty ? c0 : min(c0,b.left),fuzz,maxdepth);
@@ -446,11 +443,11 @@ void drawBezierTriangle::ratio(const double* t, pair &b,
 bool drawBezierTriangle::write(prcfile *out, unsigned int *, double, 
                                groupsmap&)
 {
-  if(invisible)
+  if(invisible || primitive)
     return true;
 
   RGBAColour Black(0.0,0.0,0.0,diffuse.A);
-  PRCmaterial m(Black,diffuse,emissive,specular,opacity,PRCshininess);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
   
   static const double third=1.0/3.0;
   static const double third2=2.0/3.0;
@@ -470,7 +467,7 @@ bool drawBezierTriangle::write(prcfile *out, unsigned int *, double,
 bool drawBezierTriangle::write(jsfile *out)
 {
 #ifdef HAVE_LIBGLM
-  if(invisible)
+  if(invisible || primitive)
     return true;
 
   if(billboard) {
@@ -480,11 +477,13 @@ bool drawBezierTriangle::write(jsfile *out)
   
   setcolors(colors,diffuse,emissive,specular,shininess,metallic,fresnel0,out);
   
+  out->precision(digits);
   if(straight) {
     triple Controls[]={controls[0],controls[6],controls[9]};
     out->addPatch(Controls,3,Min,Max,colors,3);
   } else
     out->addPatch(controls,10,Min,Max,colors,3);
+  out->precision(getSetting<Int>("digits"));
                     
 #endif  
   return true;
@@ -498,6 +497,17 @@ void drawBezierTriangle::render(double size2, const triple& b, const triple& B,
   transparent=colors ? colors[0].A+colors[1].A+colors[2].A < 3.0 :
     diffuse.A < 1.0;
   
+  setcolors(colors,diffuse,emissive,specular,shininess,metallic,fresnel0);
+  
+  if(transparent)
+    setMaterial(transparentData,drawTransparent);
+  else {
+    if(colors)
+      setMaterial(colorData,drawColor);
+    else
+      setMaterial(materialData,drawMaterial);
+  }
+
   bool offscreen;
   if(billboard) {
     drawElement::centerIndex=centerIndex;
@@ -509,20 +519,10 @@ void drawBezierTriangle::render(double size2, const triple& b, const triple& B,
   if(offscreen) { // Fully offscreen
     S.Onscreen=false;
     S.data.clear();
+    S.notRendered();
     return;
   }
 
-  setcolors(colors,diffuse,emissive,specular,shininess,metallic,fresnel0);
-  
-  if(transparent)
-    setMaterial(transparentData,drawTransparent);
-  else {
-    if(colors)
-      setMaterial(colorData,drawColor);
-    else
-      setMaterial(materialData,drawMaterial);
-  }
-
   triple *Controls;
   triple Controls0[10];
   if(billboard) {
@@ -573,7 +573,7 @@ bool drawNurbs::write(prcfile *out, unsigned int *, double, groupsmap&)
     return true;
 
   RGBAColour Black(0.0,0.0,0.0,diffuse.A);
-  PRCmaterial m(Black,diffuse,emissive,specular,opacity,PRCshininess);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
   out->addSurface(udegree,vdegree,nu,nv,controls,uknots,vknots,m,weights);
   
   return true;
@@ -676,17 +676,13 @@ void drawNurbs::render(double size2, const triple& b, const triple& B,
 // TODO: implement NURBS renderer
 }
 
-void drawSphere::P(triple& t, double x, double y, double z)
+void drawPRC::P(triple& t, double x, double y, double z)
 {
-  if(half) {
-    double temp=z; z=x; x=-temp;
-  }
-  
   if(T == NULL) {
     t=triple(x,y,z);
     return;
   }
-  
+
   double f=T[12]*x+T[13]*y+T[14]*z+T[15];
   if(f == 0.0) run::dividebyzero();
   f=1.0/f;
@@ -695,6 +691,14 @@ void drawSphere::P(triple& t, double x, double y, double z)
            (T[8]*x+T[9]*y+T[10]*z+T[11])*f);
 }
 
+void drawSphere::P(triple& t, double x, double y, double z)
+{
+  if(half) {
+    double temp=z; z=x; x=-temp;
+  }
+  drawPRC::P(t,x,y,z);
+}
+
 bool drawSphere::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible)
@@ -759,6 +763,31 @@ bool drawSphere::write(prcfile *out, unsigned int *, double, groupsmap&)
   return true;
 }
 
+bool drawSphere::write(jsfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  drawElement::centerIndex=0;
+
+  setcolors(false,diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  triple O,E;
+  P(E,1.0,0.0,0.0);
+  P(O,0.0,0.0,0.0);
+  triple X=E-O;
+  double r=length(X);
+
+  if(half)
+    out->addSphere(O,r,half,X.polar(false),X.azimuth(false));
+  else
+    out->addSphere(O,r);
+
+#endif  
+  return true;
+}
+
 bool drawCylinder::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible)
@@ -772,6 +801,31 @@ bool drawCylinder::write(prcfile *out, unsigned int *, double, groupsmap&)
   return true;
 }
   
+bool drawCylinder::write(jsfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  drawElement::centerIndex=0;
+  
+  setcolors(false,diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  triple E,H,O;
+  P(E,1.0,0.0,0.0);
+  P(H,0.0,0.0,1.0);
+  P(O,0.0,0.0,0.0);
+  triple X=E-O;
+  triple Z=H-O;
+  double r=length(X);
+  double h=length(Z);
+  
+  out->addCylinder(O,r,h,Z.polar(false),Z.azimuth(false),core);
+
+#endif  
+  return true;
+}
+  
 bool drawDisk::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible)
@@ -785,54 +839,53 @@ bool drawDisk::write(prcfile *out, unsigned int *, double, groupsmap&)
   return true;
 }
   
-bool drawTube::write(prcfile *out, unsigned int *, double, groupsmap&)
+bool drawDisk::write(jsfile *out)
 {
+#ifdef HAVE_LIBGLM
   if(invisible)
     return true;
 
-  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
-  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
+  drawElement::centerIndex=0;
   
-  Int n=center.length();
+  setcolors(false,diffuse,emissive,specular,shininess,metallic,fresnel0,out);
   
-  if(center.piecewisestraight()) {
-    triple *centerControls=new(UseGC) triple[n+1];
-    for(Int i=0; i <= n; ++i)
-      centerControls[i]=center.point(i);
-    size_t N=n+1;
-    triple *controls=new(UseGC) triple[N];
-    for(Int i=0; i <= n; ++i)
-      controls[i]=g.point(i);
-    out->addTube(N,centerControls,controls,true,m);
-  } else {
-    size_t N=3*n+1;
-    triple *centerControls=new(UseGC) triple[N];
-    centerControls[0]=center.point((Int) 0);
-    centerControls[1]=center.postcontrol((Int) 0);
-    size_t k=1;
-    for(Int i=1; i < n; ++i) {
-      centerControls[++k]=center.precontrol(i);
-      centerControls[++k]=center.point(i);
-      centerControls[++k]=center.postcontrol(i);
-    }
-    centerControls[++k]=center.precontrol(n);
-    centerControls[++k]=center.point(n);
-    
-    triple *controls=new(UseGC) triple[N];
-    controls[0]=g.point((Int) 0);
-    controls[1]=g.postcontrol((Int) 0);
-    k=1;
-    for(Int i=1; i < n; ++i) {
-      controls[++k]=g.precontrol(i);
-      controls[++k]=g.point(i);
-      controls[++k]=g.postcontrol(i);
-    }
-    controls[++k]=g.precontrol(n);
-    controls[++k]=g.point(n);
-    
-    out->addTube(N,centerControls,controls,false,m);
-  }
-      
+  triple E,H,O;
+  P(E,1.0,0.0,0.0);
+  P(H,0.0,0.0,1.0);
+  P(O,0.0,0.0,0.0);
+  triple X=E-O;
+  triple Z=H-O;
+  double r=length(X);
+  
+  out->addDisk(O,r,Z.polar(false),Z.azimuth(false));
+
+#endif
+  return true;
+}
+  
+bool drawTube::write(jsfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  drawElement::centerIndex=0;
+
+  setcolors(false,diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  bbox3 b;
+  b.add(T*m);
+  b.add(T*triple(m.getx(),m.gety(),M.getz()));
+  b.add(T*triple(m.getx(),M.gety(),m.getz()));
+  b.add(T*triple(m.getx(),M.gety(),M.getz()));
+  b.add(T*triple(M.getx(),m.gety(),m.getz()));
+  b.add(T*triple(M.getx(),m.gety(),M.getz()));
+  b.add(T*triple(M.getx(),M.gety(),m.getz()));
+  b.add(T*M);
+
+  out->addTube(g,width,b.Min(),b.Max(),core);
+
+#endif
   return true;
 }
 
@@ -891,11 +944,11 @@ bool drawTriangles::write(prcfile *out, unsigned int *, double, groupsmap&)
   if(nC) {
     const RGBAColour white(1,1,1,opacity);
     const RGBAColour black(0,0,0,opacity);
-    const PRCmaterial m(black,white,black,specular,opacity,PRCshininess);
+    const PRCmaterial m(black,white,black,specular,opacity,shininess);
     out->addTriangles(nP,P,nI,PI,m,nN,N,NI,0,NULL,NULL,nC,C,CI,0,NULL,NULL,30);
   } else {
     RGBAColour Black(0.0,0.0,0.0,diffuse.A);
-    const PRCmaterial m(Black,diffuse,emissive,specular,opacity,PRCshininess);
+    const PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
     out->addTriangles(nP,P,nI,PI,m,nN,N,NI,0,NULL,NULL,0,NULL,NULL,0,NULL,NULL,30);
   }
 
@@ -927,6 +980,7 @@ void drawTriangles::render(double size2, const triple& b,
   if(bbox2(Min,Max).offscreen()) { // Fully offscreen
     R.Onscreen=false;
     R.data.clear();
+    R.notRendered();
     return;
   }
 
