@@ -50,10 +50,8 @@ extern FILE *pipeout;
 
 inline void openpipeout() 
 {
-  int fd=settings::getSetting<Int>("outpipe");
-  if(!pipeout) {
-    if(fd >= 0) pipeout=fdopen(intcast(fd),"w");
-  }
+  int fd=intcast(settings::getSetting<Int>("outpipe"));
+  if(!pipeout && fd >= 0) pipeout=fdopen(fd,"w");
   if(!pipeout) {
     ostringstream buf;
     buf << "Cannot open outpipe " << fd;
@@ -114,12 +112,20 @@ public:
       standardEOF();
   }
   
-  void dimension(Int Nx=-1, Int Ny=-1, Int Nz=-1) {nx=Nx; ny=Ny; nz=Nz;}
-  
+  void dimension(Int Nx=-1, Int Ny=-1, Int Nz=-1) {
+    if(Nx < -1 || Ny < -1 || Nz < -1) {
+      ostringstream buf;
+      buf << "Invalid array dimensions: " << Nx << ", " << Ny << ", " << Nz;
+      reportError(buf);
+    }
+
+    nx=Nx; ny=Ny; nz=Nz;
+  }
+
   file(const string& name, bool check=true, Mode type=NOMODE, bool binary=false,
        bool closed=false) : 
     name(name), check(check), type(type), linemode(false), csvmode(false),
-    singlereal(false), singleint(true), signedint(true),
+    wordmode(false), singlereal(false), singleint(true), signedint(true),
     closed(closed), standard(name.empty()),
     binary(binary), nullfield(false), whitespace("") {dimension();}
   
@@ -249,7 +255,7 @@ public:
 
 class opipe : public file {
 public:
-  opipe(const string& name) : file(name,false,OPIPE) {}
+  opipe(const string& name) : file(name,false,OPIPE) {standard=false;}
 
   void open() {
     openpipeout();
@@ -425,7 +431,8 @@ public:
     ifile(name,comment,true,UPDATE,std::ios::in | std::ios::out) {}
 
   Int precision(Int p) {
-    return p == 0 ? stream->precision() : stream->precision(p);
+    return p == 0 ? stream->precision(settings::getSetting<Int>("digits")) :
+      stream->precision(p);
   }
   void flush() {if(fstream) fstream->flush();}
   
@@ -465,6 +472,7 @@ public:
     } else {
       name=outpath(name);
       stream=fstream=new std::ofstream(name.c_str(),mode | std::ios::trunc);
+      stream->precision(settings::getSetting<Int>("digits"));
       index=processData().ofile.add(fstream);
       Check();
     }
@@ -485,7 +493,8 @@ public:
   }
   void clear() {stream->clear();}
   Int precision(Int p) {
-    return p == 0 ? stream->precision() : stream->precision(p);
+    return p == 0 ? stream->precision(settings::getSetting<Int>("digits")) :
+      stream->precision(p);
   }
   void flush() {stream->flush();}
   
@@ -503,6 +512,9 @@ public:
       return 0;
   }
   
+  bool enabled() {return !standard || settings::verbose > 1 || 
+      interact::interactive || !settings::getSetting<bool>("quiet");}
+  
   void write(bool val) {*stream << (val ? "true " : "false ");}
   void write(Int val) {*stream << val;}
   void write(double val) {*stream << val;}
