@@ -2,17 +2,11 @@
 #define JSFILE_H
 
 #include <fstream>
-#include <glm/glm.hpp>
-
 #include "common.h"
 #include "triple.h"
 #include "locate.h"
 #include "prcfile.h"
 
-namespace gl {
-extern glm::mat4 projViewMat;
-}
-
 namespace camp {
 
 class jsfile {
@@ -20,16 +14,26 @@ class jsfile {
   
 public:  
   jsfile() {}
-  ~jsfile();
+  ~jsfile() {}
   
+  void precision(int digits) {out.precision(digits);}
+
   void open(string name);
-  void copy(string name);
+  void header(string name);
+  void meta(string name, bool scalable=true);
+  void comment(string name);
+  void finish(string name);
+  void footer(string name);
+
+  void svgtohtml(string name);
+
+  void copy(string name, bool header=false);
   
   void addColor(const prc::RGBAColour& c); 
   void addIndices(const uint32_t *I); 
     
   void addPatch(const triple* controls, size_t n, const triple& Min,
-                const triple& Max, const prc::RGBAColour *colors);
+                const triple& Max, const prc::RGBAColour *colors, size_t nc);
   
   void addCurve(const triple& z0, const triple& c0,
                 const triple& c1, const triple& z1,
@@ -48,6 +52,15 @@ public:
                     const uint32_t (*PI)[3], const uint32_t (*NI)[3],
                     const uint32_t (*CI)[3],
                     const triple& Min, const triple& Max);
+  void addSphere(const triple& center, double radius, bool half=false,
+                 const double& polar=0.0, const double& azimuth=0.0);
+  void addCylinder(const triple& center, double radius, double height,
+                   const double& polar, const double& azimuth,
+                   bool core=false);
+  void addDisk(const triple& center, double radius,
+               const double& polar=0.0, const double& azimuth=0.0);
+  void addTube(const triple *g, double width,
+               const triple& Min, const triple& Max, bool core=false);
 };
 
 } //namespace camp
