@@ -51,6 +51,7 @@ bool warn(const string& s);
 extern string systemDir;
 extern string docdir;
 extern const string dirsep;
+extern string tempdir;
   
 extern bool safe;
   
@@ -120,13 +121,12 @@ const double cm=inches/2.54;
 const double tex2ps=72.0/72.27;
 const double ps2tex=1.0/tex2ps;
 
-const string AsyGLVersion="1.00";
 const string AsyGL="webgl/asygl.js";
 const string WebGLheader="webgl/WebGLheader.html";
 const string WebGLfooter="webgl/WebGLfooter.html";
 }
 
-
 extern const char *REVISION;
+extern const char *AsyGLVersion;
 
 #endif
